from loguru import logger
import paho.mqtt.client as mqtt
import os
import sys
import ssl
import json
import base64
import struct
import psycopg2
import psycopg2.extras



class ApplicationNotFoundException (Exception): 
    def __init__(self, deviceId):
        self.deviceId = deviceId

class SensorTypeNotFoundException (Exception): 
    def __init__(self, sensorType):
        self.sensorType = sensorType

class JustIgnoreMessage (Exception):
    def __init__(self, application, device, message):
        self.application = application
        self.device = device
        self.message = message

class DbOp(object):
    def __init__(self, config):
        self.conn = None

    def __getConn(self):
        conn = psycopg2.connect()
        conn.autocommit = False
        return conn

    def getApplication(self, deviceId):
        try:
            conn = self.__getConn()
            with conn:
                with conn.cursor() as cur:
                    cur.execute("select label, sensor_type from application_t where device_id = %(deviceId)s", { 'deviceId': deviceId })
                    res = cur.fetchone()
                    if res is None:
                        raise ApplicationNotFoundException(deviceId)
                    application = {
                        'label': res[0],
                        'sensor_type': res[1]
                    }
                logger.debug(f"{application=}")
            return application
        except Exception as e:
            logger.error(f"Error getting device: {e}")
            raise ApplicationNotFoundException(deviceId)
        finally:
            if conn:
                conn.close()

    def storeMeasurement(self, measurement):
        try:
            logger.info(f"About to store {measurement}")
            conn = self.__getConn()
            with conn:
                with conn.cursor() as cur:
                    cur.execute("""
                        insert into measurement_t (time, application_name, moisture, temperature, conductivity, status, battery) 
                            values (now(), %(application_name)s, %(moisture)s, %(temperature)s, %(conductivity)s, %(status)s, %(battery)s)
                        """,
                        { 
                            'application_name': measurement['application_name'], 
                            'moisture': measurement['moisture'],
                            'temperature': measurement['temperature'],
                            'conductivity': measurement['conductivity'],
                            'status': measurement['status'],
                            'battery': measurement['battery']
                        }
                    )
        except Exception as e:
            logger.error(f"Error storing measurement: {e}")
        finally:
            if conn:
                conn.close()
    
def mqttOnConnectCallback(client, userdata, flags, rc):
    logger.info("mqtt connected")
    mqtt_in_topic = f"v3/{config['APPLICATION_TENANT']}/devices/+/up"
    client.subscribe(mqtt_in_topic)
    logger.info(f"subscribed to {mqtt_in_topic}")

def mqttOnMessageCallback(client, userdata, message):
    try:
        topic = message.topic
        payload = message.payload
        logger.debug(f"mqtt message received: {topic} -> {payload}")
        parse_payload = json.loads(payload)

        sendSetupMessage = False
        device_id = parse_payload['end_device_ids']['device_id']

        dbh = DbOp(config)
        application = dbh.getApplication(device_id)

        frame = base64.b64decode(parse_payload['uplink_message']['frm_payload'])
        f_port = parse_payload['uplink_message']['f_port']
        measurement = {}
        
        match application['sensor_type']:
            case 'LSE01':
                battery = struct.unpack('>H', frame[0:2])[0]
                moisture = struct.unpack('>H', frame[4:6])[0]
                temperature = struct.unpack('>H', frame[6:8])[0]
                conductivity = struct.unpack('>H', frame[8:10])[0]
                status = struct.unpack('?', frame[10:11])[0]
                logger.debug(f"{frame=}, {battery=}, {moisture=}, {temperature=}, {conductivity=}, {status=}")

                statusText = 'Ok'

                measurement = {
                    'application_name': application['label'],
                    'moisture': moisture / 100,
                    'temperature': temperature / 100,
                    'conductivity': conductivity,
                    'status': statusText,
                    'battery': battery / 1000
                }
            case _:
                raise SensorTypeNotFoundException(application['sensor_type'])

        logger.debug(f"{measurement=}")
        dbh.storeMeasurement(measurement)
    except JustIgnoreMessage as e:
        logger.info(f"JustIgnoreMessage: {e.application}, {e.device}, {e.message}")
    except SensorTypeNotFoundException as e:
        logger.error(f"application has unknown sensor type {e.sensorType}")
    except ApplicationNotFoundException as e:
        logger.error(f"message from unknown device {e.deviceId}")
    except Exception as e:
        logger.error(f"unable to parse message {payload}, {e}")

def mqttOnDisconnectCallback(client, userdata, rc):
    pass

logger.info("preprocess starting")



REQUIRED_CONFIG_OPTIONS = [ 
    'MQTT_LOGIN',
    'MQTT_PASSWORD',
    'MQTT_BROKER',
    'MQTT_PORT',
    'MQTT_CA',
    'APPLICATION_TENANT'
]

config = {}
for rco in REQUIRED_CONFIG_OPTIONS:
    try:
        config[rco] = os.environ[rco]
    except KeyError:
        logger.error(f"{rco} is a required config option, not available in environment")
        sys.exit(-1)



client = mqtt.Client()
client.on_message = mqttOnMessageCallback
client.on_connect = mqttOnConnectCallback
client.on_disconnect = mqttOnDisconnectCallback
client.username_pw_set(config['MQTT_LOGIN'], config['MQTT_PASSWORD'])
client.tls_set(
    cert_reqs=ssl.CERT_REQUIRED,
    ciphers=None
)
client.connect(config["MQTT_BROKER"], int(config["MQTT_PORT"]))
#client.connect('172.16.2.16', 1883)
logger.info("mqtt loop starting")
client.loop_forever()




